/**
 * File:
 *  modbusSlave.c
 *
 * Notes:
 *  This file contains the implementation for a serial RTU modbus slave.
 *
 * Functions:
 *  addCoil         Add a coil to the slave
 *  addHoldingReg   Add a holding register to the slave
 *  addInputReg     Add an input register to the slave
 *  addStatusBit    Add a status bit to the slave
 *  calcCRC         Calculates the CRC for the message content
 *  coilState       Sets the state of a coil
 *  decodePacket    Decodes a packet received from modbus master
 *  modbusSlave     Initialise modbus slave structure
 *  packBits        Packs bits into a message buffer
 *  packRegisters   Packs registers into a message buffer
 *  setRegister     Sets the value of a holding register
 *  startTimeout    Starts the message timeout timer
 *  validAddresses  Validates an address
 *  validDiscretes  Validates a discrete
 *  validRegisters  Validates a register
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  19/06/2012 Added receivePacket function which buffers a packet
 *  21/06/2012 Implemented exception responses
 *             Implemented Timer0 to time receipt of bytes
 *  25/06/2012 Implemented startTimeout
 *  27/06/2012 Replaced individual request and response buffers with a single
 *             message buffer.  Re-wrote CRC to use lookup tables.  Improved
 *             comments.
 */
#include <built_in.h>

#include "modbusSlave.h"
// Pointers to last node in each linked list
static coilDef*    pLastCoil       = NULL;
static statusDef*  pLastStatusBit  = NULL;
static inputDef*   pLastInput      = NULL;
static holdingDef* pLastHoldingReg = NULL;
// Pointers to linked lists for each data type
static coilDef*    pCoils          = NULL;
static statusDef*  pStatusBits     = NULL;
static inputDef*   pInputs         = NULL;
static holdingDef* pHoldingRegs    = NULL;
// Flags to use when validating address ranges
static boolean blnValidStart, blnValidEnd;
// Exception code
static mbException exceptionCode;
// The modbus slave address
static byte bytSlaveAddress;
// The requested start address high byte and low byte
static byte bytStartHi, bytStartLo;
// The requested total number of items high byte and low byte
static byte bytTotalHi, bytTotalLo;
// The byte index
static byte bytIndex = 0;
// Receiver GAP set-point
static uint uintRxGapSetPt;
// Receiver GAP counter
static uint uintRxGap;
// Received packet buffer
static byte aryBuffer[MAX_RECEIVED_PACKET_LENGTH];
static byte bytRxphase;
/**
 * Function:
 *  addDiscrete
 *
 * Parameters:
 *  pDiscrete, pointer to an instance of discreteDef
 *  uintAddress, the address to assign to the discrete (base 1)
 *  pByte, pointer to the byte containing the bit to map as the discrete
 *  bBit, the bit in the byte to map to the discrete, 0 to 7
 *  pLinkList, pointer to pointer of discrete linked list
 *  pLast, pointer to pointer of last node in list
 *
 * Returns:
 *  TRUE if discrete added, FALSE if not
 */
static boolean addDiscrete(discreteDef* pDiscrete,
                           uint uintAddress,
                           ushort* pByte, byte bBit,
                           discreteDef** pLinkList,
                           discreteDef** pLast) {
  if ( !(pDiscrete && uintAddress && pByte &&
         bBit >= 0 && bBit < 8 && pLinkList && pLast) ) {
    return FALSE;
  }
  if ( *pLinkList == NULL ) {
// This is the first coil
    *pLinkList = *pLast = pDiscrete;
  } else {
// Append this coil to the linked list
    (*pLast)->pNext = pDiscrete;
// Adjust the last node
    *pLast = pDiscrete;
  }
// Ensure the discrete is set-up using the passed parameters
  pDiscrete->uintAddress = uintAddress;
  pDiscrete->pByte = pByte;
  pDiscrete->bBit  = bBit;
  pDiscrete->pNext = NULL;
}
/**
 * Function:
 *  addRegister
 *
 * Parameters:
 *  pRegister, pointer to an instance of registerDef
 *  uintAddress, the address to assign to the register (base 1)
 *  pWord, pointer to the word to map as the register
 *  pLinkList, pointer to pointer of register linked list
 *  pLast, pointer to pointer of last node in list
 *
 * Returns:
 *  TRUE if register added, FALSE if not
 */
static boolean addRegister(registerDef* pRegister,
                           uint uintAddress,
                           ushort* pWord,
                           registerDef** pLinkList,
                           registerDef** pLast) {
  if ( !(pRegister && uintAddress && pWord && pLinkList && pLast) ) {
    return FALSE;
  }
  if ( *pLinkList == NULL ) {
// This is the first coil
    *pLinkList = *pLast = pRegister;
  } else {
// Append this coil to the linked list
    (*pLast)->pNext = pRegister;
// Adjust the last node
    *pLast = pRegister;
  }
// Ensure the discrete is set-up using the passed parameters
  pRegister->uintAddress = uintAddress;
  pRegister->pWord = pWord;
  pRegister->pNext = NULL;
}
/**
 * Function:
 *  calcCRC
 *
 * Parameters:
 *  uintLength, the number of bytes in the message
 *
 * Returns:
 *  A 16bit CRC for the passed message
 */
static uint calcCRC(uint uintLength) {
  static code byte arybytHi[] = {
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
  0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
  0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
  0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
  0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
  0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
  0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
  0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
  0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
  0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
  0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40,
  0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1,
  0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
  0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
  0x80, 0x41, 0x00, 0xC1, 0x81, 0x40 };
  static code char arybytLo[] = {
  0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06,
  0x07, 0xC7, 0x05, 0xC5, 0xC4, 0x04, 0xCC, 0x0C, 0x0D, 0xCD,
  0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09,
  0x08, 0xC8, 0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A,
  0x1E, 0xDE, 0xDF, 0x1F, 0xDD, 0x1D, 0x1C, 0xDC, 0x14, 0xD4,
  0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3,
  0x11, 0xD1, 0xD0, 0x10, 0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3,
  0xF2, 0x32, 0x36, 0xF6, 0xF7, 0x37, 0xF5, 0x35, 0x34, 0xF4,
  0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A,
  0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38, 0x28, 0xE8, 0xE9, 0x29,
  0xEB, 0x2B, 0x2A, 0xEA, 0xEE, 0x2E, 0x2F, 0xEF, 0x2D, 0xED,
  0xEC, 0x2C, 0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26,
  0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 0xA0, 0x60,
  0x61, 0xA1, 0x63, 0xA3, 0xA2, 0x62, 0x66, 0xA6, 0xA7, 0x67,
  0xA5, 0x65, 0x64, 0xA4, 0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F,
  0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68,
  0x78, 0xB8, 0xB9, 0x79, 0xBB, 0x7B, 0x7A, 0xBA, 0xBE, 0x7E,
  0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 0xB4, 0x74, 0x75, 0xB5,
  0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71,
  0x70, 0xB0, 0x50, 0x90, 0x91, 0x51, 0x93, 0x53, 0x52, 0x92,
  0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54, 0x9C, 0x5C,
  0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B,
  0x99, 0x59, 0x58, 0x98, 0x88, 0x48, 0x49, 0x89, 0x4B, 0x8B,
  0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
  0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42,
  0x43, 0x83, 0x41, 0x81, 0x80, 0x40 };
  byte bytHi = 0xff, bytLo = 0xff;
  byte* pBuffer = aryBuffer;
  uint uintIdx;

  while( uintLength-- ) {
    uintIdx = bytHi ^ *pBuffer++;
    bytHi = bytLo ^ arybytHi[uintIdx];
    bytLo = arybytLo[uintIdx];
  }
// This might look odd and wrong, but it is correct!
  return (uint)(((uint)bytHi << 8) | bytLo);
}
/**
 * Function:
 *  coilState
 *
 * Parameters:
 *  uintAddr, the address of the coil to modify
 *  eState, TRUE to turn coil on, FALSE to turn coil off
 *
 * Returns:
 *  none
 */
static void coilState(uint uintAddr, boolean eState) {
// Find the first I/O point
  discreteDef* pNode;
  byte bytMask;
  for( pNode=pCoils; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    bytMask = (1 << pNode->bBit);

    if ( eState == TRUE ) {
// Turn on
      *pNode->pByte |= bytMask;
    } else {
// Turn off
      *pNode->pByte &= (0xffff ^ bytMask);
    }
    break;
  }
}
/**
 * Function:
 *  packBits
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the bits were packed into
 */
static uint packBits(discreteDef* pLL, byte* pBuffer, 
                     uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0, uintBit = 0;
  discreteDef* pNode;
  byte bytSet;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      bytSet = (*pNode->pByte & (1 << pNode->bBit)) ? 1 : 0;
      
      if ( bytSet ) {
        pBuffer[uintBytes] |= (bytSet << uintBit);
      }
      if ( ++uintBit > 7 ) {
        uintBit = 0;
        uintBytes++;
      }
    }
  }
  return uintBytes + 1;
}
/**
 * Function:
 *  packRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the registers were packed into
 */
static uint packRegisters(registerDef* pLL, byte* pBuffer,
                          uint uintStart, uint uintEnd) {
// Find the first I/O point
  uint uintBytes = 0;
  registerDef* pNode;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress >= uintStart && pNode->uintAddress <= uintEnd ) {
      pBuffer[uintBytes++] = (*pNode->pWord >> 0x08);
      pBuffer[uintBytes++] = (*pNode->pWord & 0xff);
    }
  }
  return uintBytes;
}
/**
 * Function:
 *  setRegister
 *
 * Parameters:
 *  uintAddr, the address of the register to modify
 *  bytHi, the hi-byte of the data value
 *  bytLo, the lo-byte of the data value
 *
 * Returns:
 *  none
 */
static void setRegister(uint uintAddr, byte bytHi, byte bytLo) {
// Find the first I/O point
  registerDef* pNode;
  for( pNode=pHoldingRegs; pNode!=NULL; pNode=pNode->pNext ) {
    if ( pNode->uintAddress != uintAddr  ) {
      continue;
    }
    *pNode->pWord = ((int)bytHi << 0x08) | (int)bytLo;
    break;
  }
}
/**
 * Function:
 *  startTimeout
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void startTimeout(void) {
// Disable interrupt
  TMR0ON_bit = 0;
// Set timer presets
  TMR0H = Hi(uintRxGapSetPt);
  TMR0L = Lo(uintRxGapSetPt);
// Clear the interrupt mask
  TMR0IF_bit = 0;
// Enable interrupt
  TMR0ON_bit = 1;
}
/**
 * Function:
 *  validAddresses
 *
 * Parameters:
 *  uintAddress, the address to validate against
 *  uintStart, the first address to validate
 *  uintEnd, the last address to validate
 *
 * Returns:
 *  TRUE of FALSE if both addresses are valid
 */
static boolean validAddresses(uint uintAddress, uint uintStart, uint uintEnd) {
  if ( blnValidStart == FALSE && uintStart == uintAddress ) {
    blnValidStart = TRUE;
  }
  if ( blnValidEnd == FALSE && uintEnd == uintAddress ) {
    blnValidEnd = TRUE;
  }
  return (blnValidStart == TRUE && blnValidEnd == TRUE);
}
/**
 * Function:
 *  validDiscretes
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the discrete items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validDiscretes(discreteDef* pLL, uint uintStart, uint uintEnd) {
  discreteDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  validRegisters
 *
 * Parameters:
 *  pLL, a pointer to the linked list containing the register items
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  TRUE if the start and end addresses are valid, FALSE if not
 */
static boolean validRegisters(registerDef* pLL, uint uintStart, uint uintEnd) {
  registerDef* pNode;
  blnValidStart = blnValidEnd = FALSE;
  for( pNode=pLL; pNode!=NULL; pNode=pNode->pNext ) {
    if ( validAddresses(pNode->uintAddress, uintStart, uintEnd) == TRUE ) {
      return TRUE;
    }
  }
  return FALSE;
}
/**
 * Function:
 *  addCoil
 *
 * Parameters:
 *  pCoil, pointer to an instance of coilDef
 *  uintAddress, the address to assign to the coil (base 1)
 *  pByte, pointer to the byte containing the bit to map as the coil
 *  bBit, the bit in the byte to map to the coil, 0 to 7
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addCoil(coilDef* pCoil, uint uintAddress, ushort* pByte, byte bBit) {
  return addDiscrete(pCoil, uintAddress, pByte, bBit, &pCoils, &pLastCoil);
}
/**
 * Function:
 *  addHoldingReg
 *
 * Parameters:
 *  pHolding, pointer to an instance of holdingDef
 *  uintAddress, the address to assign to the holding register (base 1)
 *  pWord, pointer to the word to use for the holding register
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addHoldingReg(holdingDef* pHolding, uint uintAddress, ushort* pWord) {
  return addRegister(pHolding, uintAddress, pWord, &pHoldingRegs, 
                     &pLastHoldingReg);
}
/**
 * Function:
 *  addInputReg
 *
 * Parameters:
 *  pInput, pointer to an instance of inputDef
 *  uintAddress, the address to assign to the input register (base 1)
 *  pWord, pointer to the word to use for the input register
 *
 * Returns:
 *  TRUE if coil added, FALSE if not
 */
boolean addInputReg(inputDef* pInput, uint uintAddress, ushort* pWord) {
  return addRegister(pInput, uintAddress, pWord, &pInputs, &pLastInput);
}
/**
 * Function:
 *  addStatusBit
 *
 * Parameters:
 *  pStatus, pointer to an instance of statusDef
 *  uintAddress, the address to assign to the status bit (base 1)
 *  pByte, pointer to the byte containing the bit to map as the status bit
 *  bBit, the bit in the byte to map to the status bit, 0 to 7
 *
 * Returns:
 *  TRUE if status bit added, FALSE if not
 */
boolean addStatusBit(statusDef* pStatus, uint uintAddress, ushort* pByte, 
                     byte bBit) {
  return addDiscrete(pStatus, uintAddress, pByte, bBit, &pStatusBits, 
                     &pLastStatusBit);
}
/**
 * Function:
 *  decodePacket
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
void decodePacket(void) {
  uint uintCRCpos = 0, uintCRC;

  if ( TMR0IF_bit ) {
    bytRxphase++;
    TMR0ON_bit = 0;
// Clear the interrupt mask
    TMR0IF_bit = 0;
  }
  if( RCIF_bit ) {
    byte bytTemp;
    bytTemp = UART_Read();
// Silent interval passed? 
// If passed, is it our address? Enough space in buffer?
    if ( bytRxphase == 0 ||
         bytIndex >= MAX_RECEIVED_PACKET_LENGTH ||
         (bytRxphase == 1 && bytTemp != bytSlaveAddress) ) {
      bytIndex = 0;              // ...no, wait for end of current message
      bytRxphase = 0;
      startTimeout();
      return;
    }
    if ( bytIndex == 0) {
// Our address received, bytRxphase=2
      bytRxphase++;
    }
// Store the data in the buffer
    aryBuffer[bytIndex++] = (byte)bytTemp;
// Set inter-frame interval
    startTimeout();
  }
  if ( bytRxphase < 3 ) {
// Nothing more to do
    return;
  }
// Set the expected position of the CRC
  uintCRCpos = 6;
// Is the function supported?
  switch( aryBuffer[1] ) {
  case READ_COILS:
  case FORCE_SINGLE_COIL:
  case FORCE_MULTIPLE_COILS:
    if ( pCoils == NULL ) {
// No coils defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_STATUS_INPUTS:
    if ( pStatusBits == NULL ) {
// No status inputs defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_HOLDING_REGISTERS:
  case PRESET_SINGLE_REGISTER:
  case PRESET_MULTIPLE_REGISTERS:
    if ( pHoldingRegs == NULL ) {
// No holding registers defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  case READ_INPUT_REGISTERS:
    if ( pInputs == NULL ) {
// No input registers defined!
      exceptionCode = ILLEGAL_DATA_ADDRESS;
    }
    break;
  default:
// No, function not supported!
    exceptionCode = ILLEGAL_FUNCTION;
    break;
  }
  bytStartHi = aryBuffer[2];
  bytStartLo = aryBuffer[3];
  bytTotalHi = aryBuffer[4];
  bytTotalLo = aryBuffer[5];

  if ( aryBuffer[1] == FORCE_MULTIPLE_COILS ||
       aryBuffer[1] == PRESET_MULTIPLE_REGISTERS )  {
    uintCRCpos += 1 + aryBuffer[6];
  }
// Calculate the packet CRC
  uintCRC = calcCRC(uintCRCpos);
// Compare the calculate checksum with the received checksum
  if ( (uintCRC >> 0x08) == aryBuffer[uintCRCpos] &&
       (uintCRC & 0xff)  == aryBuffer[uintCRCpos + 1] ) {
// CRC ok
    uint uintSAddr, uintEAddr, uintItemCount, uintLength = 6;
    uint uintOffset, uintAddr, uintBit;
    discreteDef* pDiscretes;
    registerDef* pRegisters;
// Get the start address
    uintSAddr = (((uint)bytStartHi << 0x08) | bytStartLo) + 1;
// How many items have been requested?
    uintItemCount = ((uint)bytTotalHi << 0x08) | bytTotalLo;
// Whats the last address?
    uintEAddr = uintSAddr + uintItemCount - 1;
// What was the requested function code?
    switch( aryBuffer[1] ) {
    case READ_COILS:
    case READ_STATUS_INPUTS:
      if ( aryBuffer[1] == READ_COILS ) {
        pDiscretes = pCoils;
      } else {
        pDiscretes = pStatusBits;
      }
      if ( validDiscretes(pDiscretes, uintSAddr, uintEAddr) == FALSE ) {
        exceptionCode = ILLEGAL_DATA_ADDRESS;
// Exception, address does not exist
      } else {
        aryBuffer[2] = packBits(pDiscretes,
                                &aryBuffer[3],
                                uintSAddr, uintEAddr);
        uintLength = aryBuffer[2] + 3;
      }
      break;
    case READ_HOLDING_REGISTERS:
    case READ_INPUT_REGISTERS:
    case READ_STATUS_INPUTS:
      if ( aryBuffer[1] == READ_INPUT_REGISTERS ) {
        pRegisters = pInputs;
      } else {
        pRegisters = pHoldingRegs;
      }
      if ( validRegisters(pRegisters, uintSAddr, uintEAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        aryBuffer[2] = packRegisters(pRegisters,
                                     &aryBuffer[3],
                                     uintSAddr, uintEAddr);
        uintLength = aryBuffer[2] + 3;
      }
      break;
    case FORCE_SINGLE_COIL:
      if ( validDiscretes(pCoils, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        if ( aryBuffer[4] == 0xff && aryBuffer[5] == 0x0 ) {
// Force coil on
          coilState(uintSAddr, TRUE);
        } else if ( aryBuffer[4] == 0x0 && aryBuffer[5] == 0x0 ) {
// Force coil off
          coilState(uintSAddr, FALSE);
        }
      }
      break;
    case PRESET_SINGLE_REGISTER:
      if ( validRegisters(pHoldingRegs, uintSAddr, uintSAddr) == FALSE ) {
// Exception, address does not exist
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else {
        setRegister(uintSAddr, aryBuffer[4], aryBuffer[5]);
      }
      break;
    case FORCE_MULTIPLE_COILS:
      if ( uintItemCount > MAX_DISCRETES_IN_FC15 ||
           uintItemCount > (aryBuffer[6] * 8) ) {
// Exception, either to many items or the coil count doesn't match byte count
        exceptionCode = ILLEGAL_DATA_ADDRESS;
        break;
      }
    case PRESET_MULTIPLE_REGISTERS:
      if ( (aryBuffer[1] == PRESET_MULTIPLE_REGISTERS &&
            uintItemCount > MAX_REGISTERS_IN_FC16) ) {
// Exception, either to many items or the coil count doesn't match byte count
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      } else if ( validDiscretes(pCoils,
                                 uintSAddr, uintEAddr) == FALSE ) {
        aryBuffer[1] |= EXCEPTION_FLAG;
      } else {
        uintOffset = 0;
        uintAddr = uintSAddr;

        while( uintAddr<=uintEAddr ) {
          if ( aryBuffer[1] == PRESET_MULTIPLE_REGISTERS ) {
            setRegister(uintAddr,
                        aryBuffer[7 + uintOffset],
                        aryBuffer[8 + uintOffset]);
            uintOffset += 2;
          } else {
            for(uintBit=0; uintBit<8; uintBit++, uintAddr++ ) {
              coilState(uintAddr,
                        aryBuffer[7 + uintOffset] & (1 << uintBit));
            }
            uintOffset++;
          }
          uintAddr++;
        }
      }
      break;
    }
    if( exceptionCode != NO_EXCEPTION ) {
      aryBuffer[1] |= EXCEPTION_FLAG;
      aryBuffer[2] = (byte)exceptionCode;
      exceptionCode = NO_EXCEPTION;
      uintLength = 3;
    }
    if ( uintLength > 0 ) {
// Calculate CRC for response
      uintCRC = calcCRC(uintLength);
      aryBuffer[uintLength] = (uintCRC >> 0x08);
      uintLength++;
      aryBuffer[uintLength] = (uintCRC & 0xff);
      uintLength++;
// Write the response
      for( bytIndex=0; bytIndex<uintLength; bytIndex++ ) {
        UART1_Write(aryBuffer[bytIndex]);
      }
// Wait for all data to be transmitted
      while( !TXSTA.TRMT ) ;
// Set silent interval
      bytIndex = 0;
      bytRxphase = 0;
      startTimeout();
    }
  }
}
/**
 * Function:
 * modbusSlave
 *
 * Parameters:
 *  eBaud, see modbusSlave.h for options
 *  bytAddress, the modbus slave address 1 to 255
 *
 * Returns:
 *  0 if ok, -1 if error
 */
int modbusSlave(baudRate eBaud, const byte bytAddress) {
  long lngTimeoutPreset;
  byte bytTemp;

  if ( bytAddress < 1 ) {
    return -1;
  }
// Make sure the received array is cleared
  memset(aryBuffer, 0, sizeof(aryBuffer));
// Set the slave address
  bytSlaveAddress = bytAddress;
// Initialise the Rx Gap counter
  uintRxGap = 0;
// Ensure there is no exception code
  exceptionCode = NO_EXCEPTION;
// Initialize UART module for the specified bps
  switch( eBaud ) {
  case BAUD_1200:
    UART1_Init(BAUD_1200 * 100);
    break;
  case BAUD_2400:
    UART1_Init(BAUD_2400 * 100);
    break;
  case BAUD_4800:
    UART1_Init(BAUD_4800 * 100);
    break;
  case BAUD_9600:
    UART1_Init(BAUD_9600 * 100);
    break;
  case BAUD_19200:
    UART1_Init(BAUD_19200 * 100);
    break;
 case BAUD_38400:
    UART1_Init(BAUD_38400 * 100);
    break;
  case BAUD_57600:
    UART1_Init(BAUD_57600 * 100);
    break;
  case BAUD_115200:
    UART1_Init(BAUD_115200 * 100);
    break;
  default:
    return -1;
  }
// Calculate the timer preset
  lngTimeoutPreset = MODBUS_INTERPACKET_GAP * Clock_kHz() / (4 * eBaud);
// Wait for UART module to stabilize
  Delay_ms(100);
// Timer0 Registers:
// 16-Bit Mode
// Prescaler=1:1
// TMRH Preset=FC
// TMRL Preset=E0
// Freq=20,000.00Hz
// Period=0.05 ms
  T0CON.TMR0ON = 1;// Timer0 On/Off Control bit 1=Enables / 0=Stops
  T0CON.T08BIT = 0;// Timer0 8-bit/16-bit Control bit: 1=8-bit/ 0=16-bit
  T0CON.T0CS   = 0;// TMR0 Clock Source Select bit: 0=Internal
  T0CON.T0SE   = 0;// TMR0 Source Edge Select bit: 0=low/high
  T0CON.PSA    = 1;// Prescaler Assignment bit: 1=NOT assigned/bypassed
  T0CON.T0PS2  = 0;// bits 2-0  PS2:PS0: Prescaler Select bits
  T0CON.T0PS1  = 0;
  T0CON.T0PS0  = 0;
// Set prescaler if needed
  bytTemp = 0;
  while( lngTimeoutPreset > 65535 ) {
    lngTimeoutPreset >>= 1;
    bytTemp++;
  }
  if ( bytTemp > 0 && bytTemp <= 8 ) {
    T0CON.PSA = 0;   // Prescaler assigned
    T0CON = T0CON | (--bytTemp & 0x07);
  }
  uintRxGapSetPt = (uint)-lngTimeoutPreset;
// Clear errors, Rx buffer & Rx phase indicator
  bytRxphase = RCREG;
  bytRxphase = RCREG;
  RCSTA.CREN = 0;
  RCSTA.CREN = 1;
  bytRxphase = 0;
  bytIndex = 0;
  RCIF_bit = 0;
// Start silent interval in case bus is active
  startTimeout();
  TMR0IE_bit = 1;
// Turn ON interrupt on UART1 receive
  RC1IE_bit  = 1;
// Enable peripheral interrupts
  PEIE_bit   = 1;
// Enable GLOBAL interrupts
  GIE_bit    = 1;
  return 0;
}