/**
 * File:
 *  modbusSlave.h
 *
 * Notes:
 *  This file contains the prototype for a serial RTU modbus slave.
 *
 * Usage:
 *  modbusSlaveDef slave;
 *  modbusSlave(&slave, BAUD_19200, 1);
 *  while(1) {
 *    listenToMaster(&slave);
 *  }
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  27/06/2012 Corrected MAX_DISCRETES_IN_FC15 and MAX_DISCRETES_IN_FC16
 *             Removed constants for frequency and timer presets.
 */
#ifndef MODBUS_SLAVE_H
  #define MODBUS_SLAVE_H
  
  #include "types.h"
// Constants
  #define EXCEPTION_FLAG                   0x80
  #define MAX_RECEIVED_PACKET_LENGTH       255
  #define MAX_RESPONSE_PACKET_LENGTH       MAX_RECEIVED_PACKET_LENGTH
  #define MAX_DISCRETES_IN_FC15            1976
  #define MAX_REGISTERS_IN_FC16            123
  #define MAX_MESSAGE_LENGTH               (255 + 9)
  #define MODBUS_INTERPACKET_GAP           495 // 4.5 * 110
// Type definition for storing Hi-byte and Lo-byte pair
  typedef struct {
    byte bytHi;
    byte bytLo;
  } hiloDef;
// Supported Baud rates / 100
  typedef enum eBaudRates {
    BAUD_1200   = 12,
    BAUD_2400   = 24,
    BAUD_4800   = 48,
    BAUD_9600   = 96, 
    BAUD_19200  = 192,
    BAUD_38400  = 384,
    BAUD_57600  = 576,
    BAUD_115200 = 1152
  } baudRate;
// Modbus function codes
  typedef enum eModbusFunctions {
    READ_COILS                = 1,
    READ_STATUS_INPUTS        = 2,
    READ_HOLDING_REGISTERS    = 3,
    READ_INPUT_REGISTERS      = 4,
    FORCE_SINGLE_COIL         = 5,
    PRESET_SINGLE_REGISTER    = 6,
    FORCE_MULTIPLE_COILS      = 15,
    PRESET_MULTIPLE_REGISTERS = 16
  } mbFunction;
// Modbus exception codes
  typedef enum eModbusExceptions {
    NO_EXCEPTION              = 0,
    ILLEGAL_FUNCTION          = 1,
    ILLEGAL_DATA_ADDRESS      = 2,
    ILLEGAL_DATA_VALUE        = 3,
    SLAVE_DEVICE_FAILURE      = 4,
    ACKNOWLEDGE               = 5,
    SLAVE_DEVICE_BUSY         = 6,
    NEGATIVE_ACKNOWLEDGE      = 7,
    MEMORY_PARITY_ERROR       = 8
  } mbException;
// Type defintions
  typedef struct _discrete {
// The address of the coil, 1 to 65535
    uint          uintAddress;
// A pointer to the byte containing the coil
    ushort*       pByte;
// The bit 0 to 7
    byte          bBit;
// Link to next coil
    struct _discrete* pNext;
  } discreteDef;

  typedef discreteDef coilDef;
  typedef discreteDef statusDef;

// Request data item packet structure
  typedef struct {
    byte    bytSlave;
    byte    bytFunction;
    hiloDef addr;
    hiloDef points;
    hiloDef crc;
  } reqDataDef;
  
  typedef reqDataDef reqCoilsDef;
  typedef reqDataDef reqStatusDef;
  typedef reqDataDef reqHoldReqDef;
  typedef reqDataDef reqInputsDef;

  typedef struct _register {
// The address of the register, 1 to 65535
    uint          uintAddress;
// A pointer to the word representing the register
    ushort*       pWord;
// Link to next register
    struct _register* pNext;
  } registerDef;

  typedef registerDef inputDef;
  typedef registerDef holdingDef;
// Prototypes
  boolean addCoil(coilDef* pCoil,
                  uint uintAddress, ushort* pByte, byte bBit);
  boolean addHoldingReg(holdingDef* pHolding,
                  uint uintAddress, ushort* pWord);
  boolean addInputReg(inputDef* pInput,
                  uint uintAddress, ushort* pWord);
  boolean addStatusBit(statusDef* pStatus,
                  uint uintAddress, ushort* pByte, byte bBit);
  void decodePacket(void);
  int  modbusSlave(baudRate eBaud, const byte bytAddress);
#endif